import requests
from io import IOBase
import re
import json

from dwollav2.response import Response
from dwollav2.version import version


def _items_or_iteritems(o):
    try:
        return o.iteritems()
    except:
        return o.items()


def _is_a_file(o):
    try:
        return isinstance(o, file) or isinstance(o, IOBase)
    except NameError as e:
        return isinstance(o, IOBase)


def _contains_file(o):
    if isinstance(o, dict):
        for k, v in _items_or_iteritems(o):
            if _contains_file(v):
                return True
        return False
    elif isinstance(o, tuple) or isinstance(o, list):
        for v in o:
            if _contains_file(v):
                return True
        return False
    else:
        return _is_a_file(o)


def token_for(_client):
    class Token:
        def __init__(self, opts=None, **kwargs):
            opts = kwargs if opts is None else opts
            self.access_token = opts.get('access_token')
            self.refresh_token = opts.get('refresh_token')
            self.expires_in = opts.get('expires_in')
            self.scope = opts.get('scope')
            self.app_id = opts.get('app_id')
            self.account_id = opts.get('account_id')

            self._session = requests.session()
            self._session.headers.update({
                'accept': 'application/vnd.dwolla.v1.hal+json',
                'user-agent': 'dwolla-v2-python %s' % version,
                'authorization': 'Bearer %s' % self.access_token
            })

        def post(self, url, body=None, headers={}, **kwargs):
            body = kwargs if body is None else body
            requests = _client.requests.copy()
            headers = self._merge_dicts(
                requests.pop('headers', {}), headers)
            if _contains_file(body):
                files = [(k, v) for k, v in _items_or_iteritems(
                    body) if _contains_file(v)]
                data = [(k, v) for k, v in _items_or_iteritems(
                    body) if not _contains_file(v)]
                return Response(self._session.post(self._full_url(url), headers=headers, files=files, data=data, **requests))
            else:
                return Response(self._session.post(
                    self._full_url(url),
                    headers=self._merge_dicts(
                        {'content-type': 'application/json'}, headers),
                    data=json.dumps(body, sort_keys=True, indent=2),
                    **requests))

        def get(self, url, params=None, headers={}, **kwargs):
            params = kwargs if params is None else params
            requests = _client.requests.copy()
            headers = self._merge_dicts(
                requests.pop('headers', {}), headers)
            return Response(self._session.get(self._full_url(url), headers=headers, params=params, **requests))

        def delete(self, url, params=None, headers={}):
            requests = _client.requests.copy()
            headers = self._merge_dicts(
                requests.pop('headers', {}), headers)
            return Response(self._session.delete(self._full_url(url), headers=headers, params=params, **requests))

        def _full_url(self, path):
            if isinstance(path, dict):
                path = path['_links']['self']['href']

            if path.startswith(_client.api_url):
                return path
            elif path.startswith('/'):
                return _client.api_url + path
            else:
                path = re.sub(r'^https?://[^/]*/', '', path)
                return "%s/%s" % (_client.api_url, path)

        def _merge_dicts(self, x, y):
            z = x.copy()   # start with x's keys and values
            z.update(y)    # modifies z with y's keys and values & returns None
            return z

    return Token



from dwollav2.auth import auth_for
from dwollav2.token import token_for
from dwollav2.version import version

import requests


class Client:
    ENVIRONMENTS = {
        'production': {
            'auth_url':  'https://accounts.dwolla.com/auth',
            'token_url': 'https://api.dwolla.com/token',
            'api_url':   'https://api.dwolla.com'
        },
        'sandbox': {
            'auth_url':  'https://accounts-sandbox.dwolla.com/auth',
            'token_url': 'https://api-sandbox.dwolla.com/token',
            'api_url':   'https://api-sandbox.dwolla.com'
        }
    }

    def __init__(self, **kwargs):
        self.id = self.key = kwargs.get('id') or kwargs['key']
        self.secret = kwargs['secret']
        self.environment = kwargs.get('environment', 'production')
        if self.environment not in self.ENVIRONMENTS:
            raise ValueError('invalid environment')
        self.on_grant = kwargs.get('on_grant')
        self.requests = kwargs.get('requests', {})
        self.Auth = auth_for(self)
        self.Token = token_for(self)
        self._session = requests.session()
        self._session.headers.update(
            {'user-agent': 'dwolla-v2-python %s' % version})

    def auth(self, opts=None, **kwargs):
        return self.Auth(opts, **kwargs)

    def refresh_token(self, opts=None, **kwargs):
        return self.Auth.refresh(self.Token(opts, **kwargs))

    def token(self, opts=None, **kwargs):
        return self.Token(opts, **kwargs)

    @property
    def auth_url(self):
        return self.ENVIRONMENTS[self.environment]['auth_url']

    @property
    def token_url(self):
        return self.ENVIRONMENTS[self.environment]['token_url']

    @property
    def api_url(self):
        return self.ENVIRONMENTS[self.environment]['api_url']

    def Token(self, opts=None, **kwargs):
        return Token(self, opts, **kwargs)



try:
    from urllib.parse import urlencode
except ImportError:
    from urllib import urlencode

from dwollav2.error import Error


def _is_error(res):
    try:
        return 'error' in res.json()
    except:
        return True


def _request_token(client, payload):
    res = client._session.post(
        client.token_url, data=payload, **client.requests)
    if _is_error(res):
        raise Error.map(res)
    token = client.Token(res.json())
    if client.on_grant is not None:
        client.on_grant(token)
    return token


def auth_for(_client):
    class Auth:
        def __init__(self, opts=None, **kwargs):
            opts = kwargs if opts is None else opts
            self.redirect_uri = opts.get('redirect_uri')
            self.scope = opts.get('scope')
            self.state = opts.get('state')
            self.verified_account = opts.get('verified_account')
            self.dwolla_landing = opts.get('dwolla_landing')

        @property
        def url(self):
            return '%s?%s' % (_client.auth_url, urlencode(self._query()))

        def callback(self, params=None, **kwargs):
            params = kwargs if params is None else params
            if params.get('state') != self.state:
                raise ValueError('invalid state')
            if 'error' in params:
                raise Error.map(params)
            return _request_token(_client, {
                'client_id': _client.id,
                'client_secret': _client.secret,
                'grant_type': 'authorization_code',
                'code': params['code'],
                'redirect_uri': self.redirect_uri
            })

        def _query(self):
            d = {
                'response_type': 'code',
                'client_id': _client.id,
                'redirect_uri': self.redirect_uri,
                'scope': self.scope,
                'state': self.state,
                'verified_account': self.verified_account,
                'dwolla_landing': self.dwolla_landing
            }
            return dict((k, v) for k, v in iter(sorted(d.items())) if v is not None)

        @staticmethod
        def client():
            return _request_token(_client, {
                'client_id': _client.id,
                'client_secret': _client.secret,
                'grant_type': 'client_credentials'
            })

        @staticmethod
        def refresh(token):
            return _request_token(_client, {
                'client_id': _client.id,
                'client_secret': _client.secret,
                'grant_type': 'refresh_token',
                'refresh_token': token.refresh_token
            })

    return Auth
